"""Provides basic logging configuration for alfred experiments.

.. moduleauthor: Johannes Brachem <jbrachem@posteo.de>
"""

import logging
import queue
import threading
from pathlib import Path
from typing import Union


def prepare_file_handler(filepath: Union[str, Path]) -> logging.FileHandler:
    """Returns a :class:`~logging.FileHandler` and creates the necessary
    directories on the fly, if needed.

    Args:
        filepath: Absolute path to the targeted logfile.
    """
    logpath = Path(filepath)

    if not logpath.is_absolute():
        raise ValueError("Value of filepath must be an absolute path.")

    if logpath.is_dir():
        raise ValueError("Value of filepath must point to a file.")

    logdir = logpath.parent
    logdir.mkdir(exist_ok=True)

    return logging.FileHandler(str(logpath))


def prepare_alfred_formatter(exp_id: str) -> logging.Formatter:
    """Returns a :class:`~logging.Formatter` with the standard alfred
    logging format, including experiment id.
    """

    formatter = logging.Formatter(
        f"%(asctime)s - %(name)s - %(levelname)s - experiment id={exp_id} - %(message)s"
    )

    return formatter


def parse_level(level: str) -> int:
    """Parses level definitions in lower case strings and returns the
    approriate level attribute of the Python logging library.
    """
    lvl = level.upper()
    if lvl not in ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]:
        raise ValueError("log level must be debug, info, warning, error or critical")
    return getattr(logging, lvl)


class QueuedLoggingInterface:
    """Interface for handling logging in alfred3.

    This class provides access to two loggers: The base logger and
    the queue logger. The reasoning is the following:

    * We want logging messages to contain experiment information
        (exp_id and session_id) and we want logfiles to potentially be
        experiment specific.

    * The necessary information only becomes available to pages and elements
        once the page or its parent section get appended to the
        experiment.

    * So, logged messages are collected in a queue and logged as soon
        as the experiment becomes available.

    * In order to not lose information in between their queueing
        and their eventual logging with the queue logger, all messages
        get also logged immediately using the base logger to a general
        "admin" log (this is turned off through configuration
        in run.py for local experiments by default, to avoid doubling
        log messages in a single file).

    Args:
        base_logger: Name of the base logger.
        queue_logger: Name of the queue logger.

    Attributes:
        use_base_logger (bool): Indicates, whether the base logger
            should be used. Defaults to *True*, if a *base_logger* is
            defined upon initialization and to *False*, if *base_logger*
            is *None*. You can use this to turn off base logger usage.

        queue_logger (logging.Logger): The queue logger object. You
            can acces the logger this way to apply advanced
            configuration.

        session_id (str): Allows you to set a session ID that will be
            included in logged messages. (Defaults to "NA".)
    """

    def __init__(self, base_logger: str = None, queue_logger: str = None):
        """Constructor method."""
        self.use_base_logger = True if base_logger is not None else False
        self._base_logger = (
            logging.getLogger(base_logger) if base_logger is not None else None
        )
        self._base_logger_storage = None
        self._queue_logger = (
            logging.getLogger(queue_logger) if queue_logger is not None else None
        )
        self._queue = queue.Queue()
        self._level = None
        self.session_id = "NA"

    @property
    def queue_logger(self):
        return self._queue_logger

    @queue_logger.setter
    def queue_logger(self, logger):

        if self.queue_logger is not None:
            self.warning(
                "Queue logger already present. Overriding queue logger "
                f"{self.queue_logger} with {logger}."
            )

        self._queue_logger = logger

    def add_queue_logger(self, obj, module: str):
        name = self.loggername(obj, module)
        self.queue_logger = logging.getLogger(name)
        if "experiment" in obj.__dict__:
            self.session_id = obj.experiment.session_id
        elif "exp" in obj.__dict__:
            self.session_id = obj.exp.session_id
        else:
            self.session_id = obj.experiment.session_id
        self.log_queued_messages()

    def loggername(self, obj, module: str) -> str:
        """Returns a logger name for use in :class:`~alfred3.alfredlog.QueueLoggingInterface.

        The name has the following format::

            exp.exp_id.module_name.class_name.object_identifier

        The identifier is only added if the object has an attribute
        *instance_log* that is set to *True*. If the object
        has a *name* attribute, that is used as the identifier. Else,
        the object's *uid* is used.

        Args:
            obj: The object for which a logger name should be generated.
        """
        # remove "alfred3" from module name
        module_name = module.split(".")
        module_name.pop(0)

        name = []
        name.append("exp")
        if "experiment" in obj.__dict__:
            name.append(obj.experiment.exp_id)
        elif "exp" in obj.__dict__:
            name.append(obj.exp.exp_id)
        else:
            name.append(obj.experiment.exp_id)

        name.append(".".join(module_name))
        name.append(type(obj).__name__)

        if "instance_log" in obj.__dict__:

            try:
                if obj.instance_log:
                    try:
                        name.append(obj.name)
                    except AttributeError:
                        name.append(obj.uid)

            except AttributeError as e:
                self.debug(f"Suppressed exception: {e}")

        return ".".join(name)

    def _unpack_worker(self):
        while not self._queue.empty():
            level, msg = self._queue.get()
            lvl_logger = getattr(self.queue_logger, level)
            lvl_logger(msg)

    def log_queued_messages(self):
        """Logs messages in the queue using the queue logger."""
        if self.queue_logger is None:
            pass

        if self._level:
            self.queue_logger.setLevel(self._level)
        threading.Thread(target=self._unpack_worker, name="alfredlog").start()

    def setLevel(self, level: str):
        """Sets a level for the queue logger. Since the level will be
        stored, this can be done even before a queue logger is defined.
        """

        if level.upper() not in ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]:
            raise ValueError(
                "Level must be one of 'DEBUG', 'INFO', 'WARNING', 'ERROR', or"
                " 'CRITICAL'."
            )

        self._level = getattr(logging, level.upper())

        if self.queue_logger:
            self.queue_logger.setLevel(self._level)

    def _handle_msg(self, msg: str, level: str, *args, **kwargs):

        msg = f"session id={self.session_id} - " + msg

        if self.use_base_logger:
            getattr(self._base_logger, level)(msg, *args, **kwargs)

        if not self.queue_logger:
            self._queue.put((level, msg))
        else:
            logger_lvl = getattr(self.queue_logger, level)
            try:
                logger_lvl(msg, *args, **kwargs)
            except TypeError:
                logger_lvl(*args, **kwargs)

    def deactivate_base_logger(self):
        self._base_logger_storage = self._base_logger
        self._base_logger = None
        self.use_base_logger = False

    def activate_base_logger(self):
        if not self._base_logger:
            self._base_logger = self._base_logger_storage
        self.use_base_logger = True

    def log(self, level: str, msg: str, *args, **kwargs):
        self._handle_msg(msg, level, *args, **kwargs)

    def debug(self, msg: str, *args, **kwargs):
        self._handle_msg(msg, "debug", *args, **kwargs)

    def info(self, msg: str, *args, **kwargs):
        self._handle_msg(msg, "info", *args, **kwargs)

    def warning(self, msg: str, *args, **kwargs):
        self._handle_msg(msg, "warning", *args, **kwargs)

    def error(self, msg: str, *args, **kwargs):
        self._handle_msg(msg, "error", *args, **kwargs)

    def critical(self, msg: str, *args, **kwargs):
        self._handle_msg(msg, "critical", *args, **kwargs)

    def exception(self, msg: str, *args, **kwargs):
        self._handle_msg(msg, "exception", *args, **kwargs)


class BWCompatibleLogger(QueuedLoggingInterface):
    """Offers backwards compatibility for logging.

    This class accepts an *experiment* parameter in its logging methods.
    """

    def process_experiment(self, exp):
        self.queue_logger = logging.getLogger(f"exp.{exp.exp_id}.general")
        self.session_id = exp.session_id

    def log(self, level: str, msg: str, *args, **kwargs):
        for arg in args:
            try:
                self.process_experiment(arg)
                args.remove(arg)
            except AttributeError:
                pass

        try:
            exp = kwargs.pop("experiment")
            self.process_experiment(exp)
        except KeyError:
            pass

        try:
            super().log(level=level, msg=msg, *args, **kwargs)
        except TypeError:
            super().log(level=level, msg=msg)

    def debug(self, msg: str, *args, **kwargs):
        for arg in args:
            try:
                self.process_experiment(arg)
                args.remove(arg)
            except AttributeError:
                pass

        try:
            exp = kwargs.pop("experiment")
            self.process_experiment(exp)
        except KeyError:
            pass

        try:
            super().debug(msg=msg, *args, **kwargs)
        except TypeError:
            super().debug(msg=msg)

    def info(self, msg: str, *args, **kwargs):
        for arg in args:
            try:
                self.process_experiment(arg)
                args.remove(arg)
            except AttributeError:
                pass

        try:
            exp = kwargs.pop("experiment")
            self.process_experiment(exp)
        except KeyError:
            pass

        try:
            super().info(msg=msg, *args, **kwargs)
        except TypeError:
            super().info(msg=msg)

    def warning(self, msg: str, *args, **kwargs):
        for arg in args:
            try:
                self.process_experiment(arg)
                args.remove(arg)
            except AttributeError:
                pass

        try:
            exp = kwargs.pop("experiment")
            self.process_experiment(exp)
        except KeyError:
            pass

        try:
            super().warning(msg=msg, *args, **kwargs)
        except TypeError:
            super().warning(msg=msg)

    def error(self, msg: str, *args, **kwargs):
        for arg in args:
            try:
                self.process_experiment(arg)
                args.remove(arg)
            except AttributeError:
                pass

        try:
            exp = kwargs.pop("experiment")
            self.process_experiment(exp)
        except KeyError:
            pass

        try:
            super().error(msg=msg, *args, **kwargs)
        except TypeError:
            super().error(msg=msg)

    def critical(self, msg: str, *args, **kwargs):
        for arg in args:
            try:
                self.process_experiment(arg)
                args.remove(arg)
            except AttributeError:
                pass

        try:
            exp = kwargs.pop("experiment")
            self.process_experiment(exp)
        except KeyError:
            pass

        try:
            super().critical(msg=msg, *args, **kwargs)
        except TypeError:
            super().critical(msg=msg)

    def exception(self, msg: str, *args, **kwargs):
        for arg in args:
            try:
                self.process_experiment(arg)
                args.remove(arg)
            except AttributeError:
                pass

        try:
            exp = kwargs.pop("experiment")
            self.process_experiment(exp)
        except KeyError:
            pass

        try:
            super().exception(msg=msg, *args, **kwargs)
        except TypeError:
            super().exception(msg=msg)



import setuptools

with open("README.md") as fh:
    long_description = fh.read()

# Parse version from _version.py in package directory
# See https://packaging.python.org/guides/single-sourcing-package-version/#single-sourcing-the-version
version = {}
with open("src/alfred3/_version.py") as f:
    exec(f.read(), version)

setuptools.setup(
    name="alfred3",
    version=version["__version__"],
    author="Christian Treffenstädt, Johannes Brachem, Paul Wiemann",
    author_email="treffenstaedt@psych.uni-goettingen.de",
    description=(
        "A library for rapid development of dynamic and interactive online experiments"
        " in the social sciences."
    ),
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ctreffe/alfred",
    packages=setuptools.find_packages("src"),
    package_data={
        "alfred3": [
            "files/*",
            "static/css/*",
            "static/img/*",
            "static/js/*",
            "templates/*",
            "element/templates/*",
            "element/templates/html/*",
            "element/templates/js/*",
        ]
    },
    package_dir={"": "src"},
    install_requires=[
        "pymongo>=3.10",
        "cryptography>=3.4",
        "jinja2>=2.11",
        "Flask>=1.1",
        "thesmuggler>=1.0",
        "Click>=7.1",
        "emoji>=1.2,<=1.6.3",
        "cmarkgfm>=0.5",
        "requests>=2.25",
        "bleach>=3.3",
    ],
    extras_require={
        "dev": [
            "pytest",
            "pre-commit",
            "black",
            "flake8",
            "python-dotenv",
            "bs4",
            "selenium",
            "mongomock",
            "sphinx==4.5.0",
            "sphinx-book-theme==0.3.2",
            "sphinx-remove-toctrees==0.0.3",
            "sphinx-copybutton==0.5.0",
            "recommonmark",
        ]
    },
    entry_points="""
    [console_scripts]
    alfred3=alfred3.cli:cli
    """,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.7",
)



"""
Provides elements that don't fit into the other categories.

.. moduleauthor: Johannes Brachem <jbrachem@posteo.de>
"""
import time
from pathlib import Path
from typing import Union

from .._helper import inherit_kwargs
from .core import Element, InputElement, jinja_env


class Style(Element):
    """
    Adds CSS code to a page.

    CSS styling can be used to change the appearance of page or
    individual elements.

    Args:
        code: CSS code.
        url: Url to CSS code.
        path: Path to a .css file.
        priority: Controls the order in which CSS files are placed on a
            page. Lower numbers are included first. Can be useful, if
            you have trouble with overriding rules. In everyday use,
            it's fine to stick with the default.

    Notes:
        A style is added to a specific page, and thus only affects the
        layout of that page. To change the appearance of the whole
        experiment, you can define your styles in a .css file in your
        experiment directory and reference it in the *config.conf* in
        the option *style* of the section *layout*.

    See Also:
        * How to reference a CSS file in the *config.conf*
        * See :attr:`.Element.css_class_element` and
          :attr:`.Element.css_class_container` for information on
          element CSS classes and IDs.
        * The method :meth:`.Element.add_css` can be used to add CSS
          to a specific element.

    Examples:
        Minimal example, turning the color of a specific text element
        red. The element is selected by its id::

            import alfred3 as al
            exp = al.Experiment()

            @exp.member
            class HelloWorld(al.Page):
                name = "hello_world"

                def on_exp_access(self):
                    self += al.Text("Element 1", name="test_el")
                    self += al.Style(code="#test_el {color: red;}")

        Minimal example, turning the color of all text elements on a
        specific page red. The elements are selected by their class::

            import alfred3 as al
            exp = al.Experiment()

            @exp.member
            class HelloWorld(al.Page):
                name = "hello_world"

                def on_exp_access(self):
                    self += al.Style(code=".Text-element {color: red;}")

                    self += al.Text("Element 1", name="test_el1")
                    self += al.Text("Element 2", name="test_el2")


    """

    web_widget = None
    should_be_shown = False

    def __init__(
        self, code: str = None, url: str = None, path: str = None, priority: int = 10
    ):
        """Constructor method"""
        super().__init__()
        self.priority = priority
        self.code = code
        self.url = url

        self.path = Path(path) if path is not None else None
        self.should_be_shown = False

        code_and_path = self.code and self.path
        code_and_url = self.code and self.url
        path_and_url = self.path and self.url
        if any([code_and_path, code_and_url, path_and_url]):
            raise ValueError("You can only specify one of 'code', 'url', or 'path'.")

    @property
    def css_code(self):
        if not self.code:
            return []

        if self.path:
            p = self.experiment.subpath(self.path)

            code = p.read_text(encoding="utf-8")
            return [(self.priority, code)]
        else:
            return [(self.priority, self.code)]

    @property
    def css_urls(self):

        if self.url:
            return [(self.priority, self.url)]
        else:
            return []


class HideNavigation(Style):
    """
    Removes the forward/backward/finish navigation buttons from a page.

    See Also:

        * With :class:`.NoNavigationPage`, you can achieve the same
          result. As a rule of thump, use the HideNavigation element,
          if you want to affect the display of navigation elements
          dynamically, and the NoNavigationPage, if the page in question
          will always be displayed with navigation elements.

        * Using :class:`.JumpButtons` and :class:`.JumpList`, you can add
          custom navigation elements to a page.

        * By defining the :meth:`.Page.custom_move` method on a page,
          you can implement highly customized movement behavior.

    Examples:
        Minimal example::

            import alfred3 as al
            exp = al.Experiment()

            @exp.member
            class Demo(al.Page):
                name = "demo"

                def on_exp_access(self):
                    self += al.HideNavigation()

    """

    def __init__(self):
        """Constructor method"""
        super().__init__()
        self.code = "#page-navigation {display: none;}"


class JavaScript(Element):
    """
    Adds JavaScript to a page.

    Javascript can be used to implement dynamic behavior on the client
    side.

    Args:
        code: Javascript code.
        url: Url to Javascript code.
        path: Path to a .js file.
        priority: Controls the order in which Javascript code is placed
            on a page. Lower numbers are included first. Can be useful,
            if you have trouble with overriding code. In everyday use,
            it's fine to stick with the default.

    Notes:
        You can use the jquery API (version 3.5.1 as of 2021-01-20).

    See Also:
        * See :attr:`.Element.css_class_element` and
          :attr:`.Element.css_class_container` for information on
          element CSS classes and IDs.
        * The method :meth:`.Element.add_js` can be used to add JS
          to a specific element.

    Examples:

            In this example, we use JavaScript to create a text element
            that will lead to automatic submission of the current page,
            when a change to its input is detected::

                import alfred3 as al
                exp = al.Experiment()

                @exp.member
                class HelloWorld(al.Page):
                    name = "hello_world"

                    def on_exp_access(self):

                        js_code = '''
                        $( '#test_el' ).on('change', function() {
                            move('forward') // the move() JavaScript function is provided by alfred3.
                        })
                        '''

                        self += al.JavaScript(code=js_code)
                        self += al.TextEntry("Element 1", name="test_el")

    """

    web_widget = None
    should_be_shown = False

    def __init__(
        self, code: str = None, url: str = None, path: str = None, priority: int = 10
    ):
        """Constructor method"""
        super().__init__()
        self.priority = priority
        self.code = code
        self.url = url

        self.path = Path(path) if path is not None else None
        self.should_be_shown = False

        code_and_path = self.code and self.path
        code_and_url = self.code and self.url
        path_and_url = self.path and self.url
        if any([code_and_path, code_and_url, path_and_url]):
            raise ValueError("You can only specify one of 'code', 'url', or 'path'.")

    @property
    def js_code(self):
        if not self.code and not self.path:
            return []

        elif self.path:
            p = self.experiment.subpath(self.path)

            code = p.read_text(encoding="utf-8")
            return [(self.priority, code)]
        else:
            return [(self.priority, self.code)]

    @property
    def js_urls(self):

        if self.url:
            return [(self.priority, self.url)]
        else:
            return []


class WebExitEnabler(JavaScript):
    """
    Removes the "Do you really want to leave?" popup upon closing a page.

    By default, subjects are asked to confirm their desire to leave a
    running experiment. You can turn off this behavior by adding this
    element to a page.

    Examples:
        Minimal example::

            import alfred3 as al
            exp = al.Experiment()

            @exp.member
            class Demo(al.Page):
                name = "demo"

                def on_exp_access(self):
                    self += al.WebExitEnabler()


    """

    def __init__(self):
        """Constructor method"""
        super().__init__(code="allow_leaving();", priority=10)


class Value(InputElement):
    """
    Value elements can be used to save data without any display.

    Args:
        value: The value that you want to save.
        name: Name of the element. This should be a unique identifier.
            It will be used to identify the corresponding data in the
            final data set.

    Examples:

        Minimal example::

            import alfred3 as al
            exp = al.Experiment()

            @exp.member
            class Demo(al.Page):
                name = "demo1"

                def on_exp_access(self):
                    self += al.Value("test", name="myvalue")

    """

    def __init__(self, value: Union[str, int, float], name: str, description: str = ""):
        """Constructor method."""
        super().__init__(name=name, description=description)

        #: The value that you want to save.
        self.input: Union[str, int, float] = value
        self.should_be_shown = False


class Data(Value):
    """Alias for :class:`.Value`."""

    pass


class Callback(Element):
    """
    Triggers execution of a python function from a loaded page.

    Args:
        func (callable): Python function to be called on button click.
            The function must take zero arguments, but it can be a
            method of a class or instance.

        followup (str): What to do after the python function was called.
            Can take the following values:

            - ``refresh`` submits and reloads the current page (default),
            - ``none`` does nothing,
            - ``forward`` submits the current page and moves forward,
            - ``backward`` submits the current page and moves backward,
            - ``jump>page_name`` submits the current page and triggers
              a jump to a page with the name 'page_name'
            - ``custom`` executes custom JavaScript. If you choose this
              option, you must supply your custom JavaScript through the
              argument *custom_js*.

        submit_first (bool): If True, the current values of all input
            elements on the current page will be saved on button click,
            before *func* is called. This way, these values will be
            available in *func* through :attr:`.ExperimentSession.values`,
            if func has access to the ExperimentSession object.
            See Example 3. Defaults to True.

        delay (int): Number of seconds to wait before the callback is
            triggered. Defaults to 0.

        reset_delay (bool): If *True*, the delay will start from the
            beginning every time the page is reloaded, refreshed, or
            revisited. Defaults to *False*.

        custom_js (str): Custom JavaScript to execute after the python
            function specified in *func* was called. Only takes effect,
            if *followup* is set to 'custom'.

    Examples:

        The callback on the first page will trigger after a delay of 10
        seconds. It will print the experiment id to the terminal::

            import alfred3 as al
            exp = al.Experiment()

            @exp.member
            class Demo(al.Page):

                def on_exp_access(self):
                    self += al.Callback(func=self.print_expid, delay=10)

                def print_expid(self):
                    print(self.exp.exp_id)
    """

    web_widget = None
    should_be_shown = False

    #: Javascript template
    js_template = jinja_env.get_template("js/callback.js.j2")

    def __init__(
        self,
        func: callable,
        followup: str = "refresh",
        submit_first: bool = True,
        delay: int = 0,
        reset_delay: bool = False,
        custom_js: str = "",
    ):
        super().__init__()
        self.func = func
        self.followup = followup
        self.url = None
        self.delay_original = delay
        self.delay = None
        self.start_time = None
        self.reset_delay = reset_delay
        self.submit_first = submit_first
        self.custom_js = custom_js

    @property
    def followup(self) -> str:
        """
        str: Value of the 'followup' setting for this element.
        """
        return self._followup

    @followup.setter
    def followup(self, value):
        if value not in {
            "refresh",
            "forward",
            "backward",
            "none",
            "custom",
        } and not value.startswith("jump>"):
            raise ValueError(f"Invalid value for 'followup': {value}")
        self._followup = value

    def prepare_web_widget(self):
        # docstring inherited
        self._js_code = []
        super().prepare_web_widget()
        self.url = self.exp.ui.add_callable(self.func)

        if self.delay_original == 0:
            self.delay = self.delay_original
        elif not self.start_time:
            self.start_time = time.time()
            self.delay = self.delay_original
        elif not self.reset_delay:
            now = time.time()
            already_passed = now - self.start_time
            self.delay = self.delay_original - already_passed

        d = {}
        d["url"] = self.url
        d["followup"] = self.followup
        d["timeout"] = self.delay
        d["submit_first"] = self.submit_first
        d["custom_js"] = self.custom_js
        d["set_data_url"] = self.exp.ui.set_page_data_url

        js = self.js_template.render(d)

        self.add_js(js)


class RepeatedCallback(Element):
    """
    Triggers repeated execution of a python function from a loaded page.

    Args:
        func (callable): Python function to be called on button click.
            The function must take zero arguments, but it can be a
            method of a class or instance.

        interval (int): Number of seconds to wait between two calls to
            *func*.

        followup (str): What to do after the python function was called.
            Can take the following values:

            - ``refresh`` submits and reloads the current page (default),
            - ``none`` does nothing,
            - ``forward`` submits the current page and moves forward,
            - ``backward`` submits the current page and moves backward,
            - ``jump>page_name`` submits the current page and triggers
              a jump to a page with the name 'page_name'
            - ``custom`` executes custom JavaScript. If you choose this
              option, you must supply your custom JavaScript through the
              argument *custom_js*.

        submit_first (bool): If True, the current values of all input
            elements on the current page will be saved on button click,
            before *func* is called. This way, these values will be
            available in *func* through :attr:`.ExperimentSession.values`,
            if func has access to the ExperimentSession object.
            See Example 3. Defaults to True.

        reset_delay (bool): If *True*, the delay will start from the
            beginning every time the page is reloaded, refreshed, or
            revisited. Defaults to *False*.

        custom_js (str): Custom JavaScript to execute after the python
            function specified in *func* was called. Only takes effect,
            if *followup* is set to 'custom'.

    Examples:

        The callback on the first page will trigger every 10 seconds.
        It will print the experiment id to the terminal::

            import alfred3 as al
            exp = al.Experiment()

            @exp.member
            class Demo(al.Page):

                def on_exp_access(self):
                    self += al.RepeatedCallback(func=self.print_expid, interval=10)

                def print_expid(self):
                    print(self.exp.exp_id)
    """

    web_widget = None
    should_be_shown = False

    #: Javascript template
    js_template = jinja_env.get_template("js/repeatedcallback.js.j2")

    def __init__(
        self,
        func: callable,
        interval: int,
        followup: str = "none",
        submit_first: bool = True,
        custom_js: str = "",
    ):
        super().__init__()
        self.func = func
        self.interval = interval
        self.submit_first = submit_first
        self.url = None
        self.followup = followup
        self.custom_js = custom_js

    def prepare_web_widget(self):
        # docstring inherited
        self._js_code = []
        super().prepare_web_widget()
        self.url = self.exp.ui.add_callable(self.func)

        d = {}
        d["url"] = self.url
        d["followup"] = self.followup
        d["interval"] = self.interval
        d["submit_first"] = self.submit_first
        d["custom_js"] = self.custom_js
        d["set_data_url"] = self.exp.ui.set_page_data_url

        js = self.js_template.render(d)
        self.add_js(js)


@inherit_kwargs
class HiddenInput(InputElement):
    """
    Provides a hidden entry field.

    Args:
        {kwargs}

    Examples:
        ::

            import alfred3 as al
            exp = al.Experiment()

            @exp.member
            class Demo(al.Page):
                name = "demo"

                def on_exp_access(self):
                    self += al.HiddenInput(name="hi1", default="fixed")

    """

    base_template = jinja_env.get_template("html/EmptyElement.html.j2")
    element_template = jinja_env.get_template("html/HiddenInputElement.html.j2")



"""
Sections organize movement between pages in an experiment.

.. moduleauthor:: Johannes Brachem <jbrachem@posteo.de>, Paul Wiemann <paulwiemann@gmail.com>
"""
from random import shuffle

from ._core import ExpMember
from ._helper import inherit_kwargs
from .exceptions import AbortMove, AlfredError, ValidationError
from .page import _DefaultFinalPage, _PageCore


@inherit_kwargs
class Section(ExpMember):
    """
    The basic section, allows forward and backward movements.

    Args:
        shuffle (bool): If True, the order of all members in this
            section will be randomized every time the section is entered.
            Shuffling is not recursive, it only affects direct members
            of a section. That means, if there are subsections,
            their position in the parent section will be randomized,
            but the members within the subsection will not be affected.
            Defaults to False. Can be defined as a class attribute.
        {kwargs}

    Examples:
        Using a basic section and filling it with a page in instance
        style::

            import alfred3 as al
            exp = al.Experiment()

            exp += al.Section(name="main")

            exp.main += al.Page(title="Demo", name="DemoPage")

        Using a basic section and filling it with a page in class style::

            import alfred3 as al
            exp = al.Experiment()

            @exp.member
            class Main(al.Section): pass

            @exp.member(of_section="Main")
            class DemoPage(al.Page):
                title = "Demo"

    """

    #: Controls, whether participants can move forward from pages in
    #: this section.
    allow_forward: bool = True

    #: Controls, whether participants can move backward from *and to*
    #: pages in this section.
    allow_backward: bool = True

    #: Controls, whether participants can jump *from* pages in this
    #: section
    allow_jumpfrom: bool = True

    #: Controls, whether participants can jump *to* pages in this
    #: section.
    allow_jumpto: bool = True

    #: If *True*, pages in this section will be closed on leaving
    close_pages_on_leave: bool = False

    #: If True, the members of this section will be randomized every
    #: time the section is entered.
    shuffle: bool = False

    def __init__(
        self, title: str = None, name: str = None, shuffle: bool = None, **kwargs
    ):
        super().__init__(title=title, name=name, **kwargs)

        self._members = {}
        self._should_be_shown = True

        #: bool: Boolean flag, indicating whether the experiment session
        #: is currently operating within this section
        self.active: bool = False

        if shuffle is not None:
            self.shuffle = shuffle

        self._catch_page_hooks()

    def __contains__(self, member):
        try:
            return member.name in self.all_members or member.name in self.all_elements
        except AttributeError:
            return member in self.all_members or member in self.all_elements

    def __iadd__(self, other):
        self.append(other)
        return self

    def __getitem__(self, name):
        return self.all_members[name]

    def __setitem__(self, name, value):
        if "members" in self.__dict__ and name in self.members:
            if self.members[name] is value:
                return
            else:
                raise AlfredError(
                    f"{name} is a member of {self}. The name is reserved."
                )
        else:
            raise KeyError(
                f"{name} not found in members of {self}. You can use square bracket"
                " syntax only for changing existing pages, not for adding new ones."
                " Please use the augmented assignment operator '+=' for this purpose."
            )

    def __getattr__(self, name):
        try:
            return self.all_members[name]
        except KeyError:
            raise AttributeError(f"{self} has no attribute '{name}'.")

    def __setattr__(self, name, value):

        if "members" in self.__dict__ and name in self.members:
            if self.members[name] is value:
                return
            else:
                raise AlfredError(
                    f"{name} is a member of {self}. The name is reserved."
                )
        else:
            self.__dict__[name] = value

    def _shuffle_members(self):
        """Non-recursive shuffling of this section's members."""

        members = list(self.members.items())
        shuffle(members)
        self._members = dict(members)

    @property
    def members(self) -> dict:
        """
        Dictionary of the section's members.
        """
        return self._members

    @members.setter
    def members(self, value):
        self._members = value

    @property
    def empty(self) -> bool:
        """
        True, if there are no pages or subsections in this section.
        """
        return False if self.members else True

    @property
    def all_updated_members(self) -> dict:
        """
        Returns a dict of all members that already have exp access.
        Operates recursively, i.e. pages and subsections of subsections
        are included.
        """
        return {name: m for name, m in self.all_members.items() if m.exp is not None}

    @property
    def all_updated_pages(self) -> dict:
        """
        Returns a dict of all pages in the current section that have
        access to the experiment session. Operates recursively, i.e.
        pages in subsections are included.
        """
        pages = {}
        for name, member in self.all_updated_members.items():
            if isinstance(member, _PageCore):
                pages[name] = member

        return pages

    @property
    def all_updated_elements(self) -> dict:
        """
        Returns a dict of all elements in the current section that have
        access to the experiment session. Operates recursively, i.e.
        elements on pages in subsections are included.
        """
        elements = {}
        for page in self.all_updated_pages.values():
            elements.update(page.updated_elements)
        return elements

    @property
    def all_members(self) -> dict:
        """
        Returns a flat dict of all members in this section and its subsections.

        The order is preserved, i.e. members are listed in this dict in
        the same order in which they appear in the experiment.
        """
        members = {}

        for name, member in self.members.items():
            members[name] = member
            if isinstance(member, Section):
                members.update(member.all_members)

        return members

    @property
    def last_member(self):
        """
        Returns the last member of the current section. Can be a page
        or a subsection.
        """
        try:
            return list(self.members.values())[-1]
        except IndexError:
            return None

    @property
    def first_member(self):
        """
        Returns the first member of the current section. Can be a page
        or a subsection.
        """
        try:
            return list(self.members.values())[0]
        except IndexError:
            return None

    @property
    def first_page(self):
        """
        Returns the first page inside the current section.
        """
        try:
            return list(self.all_pages.values())[0]
        except IndexError:
            return None

    @property
    def last_page(self):
        """
        Returns the last page inside the current section.
        """
        try:
            return list(self.all_pages.values())[-1]
        except IndexError:
            return None

    @property
    def all_subsections(self) -> dict:
        """
        Returns a flat dict of all sections in this section and its subsections.

        The order is preserved, i.e. sections are listed in this dict in
        the same order in which they appear in the experiment.
        """
        subsections = {}

        for name, member in self.members.items():
            if isinstance(member, Section):
                subsections[name] = member
                subsections.update(member.all_subsections)

        return subsections

    @property
    def subsections(self) -> dict:
        """
        Returns a flat dict of all subsections in this section.

        Subsections in subsections are not included. Use
        :attr:`.all_subsections` for that purpose.
        """
        return {
            name: sec for name, sec in self.members.items() if isinstance(sec, Section)
        }

    @property
    def all_pages(self) -> dict:
        """
        Returns a flat dict of all pages in this section and its subsections.

        The order is preserved, i.e. pages are listed in this dict in
        the same order in which they appear in the experiment.
        """

        pages = {}
        for name, member in self.members.items():
            if isinstance(member, _PageCore):
                pages[name] = member
            elif isinstance(member, Section):
                pages.update(member.all_pages)

        return pages

    @property
    def all_closed_pages(self) -> dict:
        """
        Returns a flat dict of all *closed* pages in this section and its
        subsections.

        The order is preserved, i.e. pages are listed in this dict in
        the same order in which they appear in the experiment.
        """
        return {name: page for name, page in self.all_pages.items() if page.is_closed}

    @property
    def all_shown_pages(self) -> dict:
        """
        Returns a flat dict of all pages in this section and its
        subsections that have already been shown.

        The order is preserved, i.e. pages are listed in this dict in
        the same order in which they appear in the experiment.
        """
        return {
            name: page for name, page in self.all_pages.items() if page.has_been_shown
        }

    @property
    def pages(self) -> dict:
        """
        Returns a flat dict of all pages in this section.

        Pages in subsections are not included. Use :attr:`.all_pages`
        for that purpose.
        """
        return {
            name: page
            for name, page in self.members.items()
            if isinstance(page, _PageCore)
        }

    @property
    def all_elements(self) -> dict:
        """
        Returns a flat dict of all elements in this section.

        Recursive: Includes elements from pages in this section and all
        its subsections.
        """

        elements = {}
        for page in self.all_pages.values():
            elements.update(page.elements)
        return elements

    @property
    def all_input_elements(self) -> dict:
        """
        Returns a flat dict of all input elements in this section.

        Recursive: Includes elements from pages in this section and all
        its subsections.
        """

        elements = {}
        for page in self.all_pages.values():
            elements.update(page.input_elements)
        return elements

    @property
    def all_shown_input_elements(self) -> dict:
        """
        Returns a flat dict of all shown input elements in this section.

        Recursive: Includes elements from pages in this section and all
        its subsections.
        """

        elements = {}
        for page in self.all_pages.values():
            if page.has_been_shown:
                elements.update(page.input_elements)
        return elements

    @property
    def data(self) -> dict:
        """
        Returns a dictionary of user input data for all pages in this
        section and its subsections.
        """
        data = {}
        for page in self.all_pages.values():
            data.update(page.data)
        return data

    @property
    def unlinked_data(self) -> dict:
        """
        Returns a dictionary of user input data for all *unlinked* pages
        in this section and its subsections.
        """
        data = {}
        for page in self.all_pages.values():
            data.update(page.unlinked_data)

        return data

    def added_to_experiment(self, exp):
        # docstring inherited
        super().added_to_experiment(exp)
        self.log.add_queue_logger(self, __name__)
        self.on_exp_access()
        self._update_members_recursively()

    def _update_members(self):

        for member in self.members.values():
            if not member.experiment:
                member.added_to_experiment(self.exp)
            if not member.section:
                member.added_to_section(self)

    def _update_members_recursively(self):

        self._update_members()

        for member in self.members.values():
            member._update_members_recursively()

    def _generate_unset_tags_in_subtree(self):
        for i, member in enumerate(self.members.values(), start=1):

            if member.tag is None:
                member.tag = str(i)

            if isinstance(member, Section):
                member._generate_unset_tags_in_subtree()

    def append(self, *items):
        """
        Appends a variable number of pages or subsections to the section.

        In practice, it is recommended to use the augmented assignment
        operator ``+=`` instead in order to add pages or subsections.
        """
        for item in items:

            if item.name in dir(self):
                raise ValueError(f"Name of {item} is also an attribute of {self}.")

            if item.name in self.members:
                raise AlfredError(
                    f"Name '{item.name}' is already present in the experiment."
                )

            item.added_to_section(self)

            self.members[item.name] = item

            if self.experiment is not None:
                item.added_to_experiment(self.experiment)
                item._update_members_recursively()

            if not item.tag:
                item.tag = str(len(self.members) + 1)

    def on_exp_access(self):
        """
        Executed *once*, when the :class:`.ExperimentSession` becomes
        available to the section.

        .. warning:: We are currently questioning the four section hooks *on_enter*,
            *on_hand_over*, *on_resume*, and *on_leave*. Everything that you may wish
            to accomplish with these hooks can be done in page hooks. The section
            versions have some caveats that make them a bit tougher
            to use correctly. So, for the meantime, please avoid these hooks and
            use page hooks instead. The attributes :attr:`.Section.first_page`
            and :attr:`.Section.last_page` may be useful for you in this regard.

            The :meth:`.Section.on_exp_access` hook is not going anywhere, although we may
            at some point decide to introduce an alternative name for it in order
            to avoid confusion with :meth:`.Page.on_exp_access`.

        See Also:
            See "How to use hooks" for a how to on using hooks and an overview
            of available hooks.

        """
        pass

    def on_enter(self):
        """
        Executed *every time* this section is entered.

        .. warning:: We are currently questioning the four section hooks *on_enter*,
            *on_hand_over*, *on_resume*, and *on_leave*. Everything that you may wish
            to accomplish with these hooks can be done in page hooks. The section
            versions have some caveats that make them a bit tougher
            to use correctly. So, for the meantime, please avoid these hooks and
            use page hooks instead. The attributes :attr:`.Section.first_page`
            and :attr:`.Section.last_page` may be useful for you in this regard.

            The :meth:`.Section.on_exp_access` hook is not going anywhere, although we may
            at some point decide to introduce an alternative name for it in order
            to avoid confusion with :meth:`.Page.on_exp_access`.

        See Also:
            See "How to use hooks" for a how to on using hooks and an overview
            of available hooks.
        """
        pass

    def on_leave(self):
        """
        Executed *every time* this section is left.

        .. warning:: We are currently questioning the four section hooks *on_enter*,
            *on_hand_over*, *on_resume*, and *on_leave*. Everything that you may wish
            to accomplish with these hooks can be done in page hooks. The section
            versions have some caveats that make them a bit tougher
            to use correctly. So, for the meantime, please avoid these hooks and
            use page hooks instead. The attributes :attr:`.Section.first_page`
            and :attr:`.Section.last_page` may be useful for you in this regard.

            The :meth:`.Section.on_exp_access` hook is not going anywhere, although we may
            at some point decide to introduce an alternative name for it in order
            to avoid confusion with :meth:`.Page.on_exp_access`.

        See Also:
            See "How to use hooks" for a how to on using hooks and an overview
            of available hooks.
        """
        pass

    def on_resume(self):
        """
        Executed *every time* the experiment resumes from a direct subsection to this section.

        Resuming takes place, when a child section is left and the
        next page is a direct child of the parent section. Then this
        the parent section becomes the primary current section again: it
        resumes its status.

        .. warning:: We are currently questioning the four section hooks *on_enter*,
            *on_hand_over*, *on_resume*, and *on_leave*. Everything that you may wish
            to accomplish with these hooks can be done in page hooks. The section
            versions have some caveats that make them a bit tougher
            to use correctly. So, for the meantime, please avoid these hooks and
            use page hooks instead. The attributes :attr:`.Section.first_page`
            and :attr:`.Section.last_page` may be useful for you in this regard.

            The :meth:`.Section.on_exp_access` hook is not going anywhere, although we may
            at some point decide to introduce an alternative name for it in order
            to avoid confusion with :meth:`.Page.on_exp_access`.

        See Also:
            See "How to use hooks" for a how to on using hooks and an overview
            of available hooks.
        """
        pass

    def on_hand_over(self):
        """
        Executed *every time* a direct subsection of this section is entered.

        .. warning:: We are currently questioning the four section hooks *on_enter*,
            *on_hand_over*, *on_resume*, and *on_leave*. Everything that you may wish
            to accomplish with these hooks can be done in page hooks. The section
            versions have some caveats that make them a bit tougher
            to use correctly. So, for the meantime, please avoid these hooks and
            use page hooks instead. The attributes :attr:`.Section.first_page`
            and :attr:`.Section.last_page` may be useful for you in this regard.

            The :meth:`.Section.on_exp_access` hook is not going anywhere, although we may
            at some point decide to introduce an alternative name for it in order
            to avoid confusion with :meth:`.Page.on_exp_access`.

        See Also:
            See "How to use hooks" for a how to on using hooks and an overview
            of available hooks.

        """
        pass

    def _enter(self):
        self.active = True

        self.log.debug(f"Entering {self}.")
        self.on_enter()
        self._update_members()

        if self.shuffle:
            self._shuffle_members()

        if isinstance(self.first_member, Section) and not self.first_member.active:
            self._hand_over()
            self.first_member._enter()

    def _leave(self):
        self.log.debug(f"Leaving {self}.")
        self.on_leave()

        try:
            self.validate_on_leave()
        except ValidationError:
            raise AbortMove
        if self.close_pages_on_leave:
            for page in self.pages.values():
                page.close()

        if self is self.parent.last_member:
            self.parent._leave()

    def _resume(self):
        self.log.debug(f"Resuming to {self}.")
        self.on_resume()

    def _hand_over(self):
        self.log.debug(f"{self} handing over to child section.")
        self.on_hand_over()

    def _forward(self):
        pass

    def _backward(self):
        pass

    def _jumpfrom(self):
        pass

    def _jumpto(self):
        pass

    def _move(self, direction, from_page, to_page):
        """
        Conducts a section's part of moving in an alfred experiment.

        Raises:
            ValidationError: If validation of the current page fails.
        """
        if direction == "forward":
            self._forward()
        elif direction == "backward":
            self._backward()
        elif direction == "jumpto":
            self._jumpto()
        elif direction.startswith("jump"):
            self._jumpfrom()

        if to_page.section.name in self.all_members:
            self._hand_over()
        elif to_page.section is not self:
            self._leave()

        if direction.startswith("jump"):
            to_page.section._jumpto()

        if self.name in to_page.section.all_members:
            to_page.section._resume()
        elif to_page.section is not self:
            to_page.section._enter()

        if self.exp.aborted:
            raise AbortMove

    def _validate(self, direction: str):

        if direction == "forward":
            self.validate_on_forward()
        elif direction == "backward":
            self.validate_on_backward()
        elif direction.startswith("jump"):
            self.validate_on_jump()

    def validate_on_leave(self):
        """
        Validates pages and their input elements within the section.

        Can be overloaded to change the validating behavior of a derived
        section.

        Notes:
            Validation is conducted only for pages that are direct
            children of this section. Pages in subsections are not
            validated.

        Raises:
            ValidationError: If validation fails.
        """
        for page in self.pages.values():

            if not page._validate():
                raise ValidationError()

            if not page._validate_elements():
                msg = self.exp.config.get("hints", "no_input_section_validation")
                msg = msg.format(n=len(self.pages))
                self.exp.post_message(msg, level="danger")
                raise ValidationError()

    def validate_on_move(self):
        """
        Validates the current page and its elements.

        Can be overloaded to change the validating behavior of a derived
        section. By default, this validation method is called on each
        foward and backward move, as well as when participants jump
        *from* the section, but not when they jump *to* the section.

        Raises:
            ValidationError: If validation fails.

        See Also:
            Use the individual methods :meth:`.validate_on_forward`,
            :meth:`.validate_on_backward`, :meth:`.validate_on_jumpfrom`,
            and :meth:`.validate_on_jumpto` if you want even more fine-
            grained control over validation behavior.

        .. versionchanged:: 2.1.0
           Switched the default order of validation. Now, elements are
           validated first, then the page itself is validated.

        """

        if not self.exp.current_page._validate_elements():
            raise ValidationError()

        if not self.exp.current_page._validate():
            raise ValidationError()

    def validate_on_forward(self):
        """
        Called for validation on each forward move.

        Overload this method to customize validation behavior.

        See Also:
             By default, sections use :meth:`.validate_on_move` for
             validation on all kinds of moves.

        """
        self.validate_on_move()

    def validate_on_backward(self):
        """
        Called for validation on each backward move.

        Overload this method to customize validation behavior.

        See Also:
             By default, sections use :meth:`.validate_on_move` for
             validation on all kinds of moves.
        """
        self.validate_on_move()

    def validate_on_jump(self):
        """
        Called for validation on jumping from this section.

        Overload this method to customize validation behavior.

        See Also:
             By default, sections use :meth:`.validate_on_move` for
             validation on all kinds of moves.
        """
        self.validate_on_move()

    def _catch_page_hooks(self):
        """
        Raises errors, if users define page hooks on a section.
        """

        explanation = " This does not work. Remove the method to continue."

        try:
            self.on_first_show()
            msg = (
                "You tried to use the page-only hook method 'on_first_show' on the"
                f" section {self}."
            )
            raise AlfredError(msg + explanation)
        except AttributeError:
            pass

        try:
            self.on_each_show()
            msg = (
                "You tried to use the page-only hook method 'on_each_show' on the"
                f" section {self}."
            )
            raise AlfredError(msg + explanation)
        except AttributeError:
            pass

        try:
            self.on_each_hide()
            msg = (
                "You tried to use the page-only hook method 'on_each_hide' on the"
                f" section {self}."
            )
            raise AlfredError(msg + explanation)
        except AttributeError:
            pass

        try:
            self.on_first_hide()
            msg = (
                "You tried to use the page-only hook method 'on_first_hide' on the"
                f" section {self}."
            )
            raise AlfredError(msg + explanation)
        except AttributeError:
            pass


@inherit_kwargs
class RevisitSection(Section):
    """
    A section that disables all input elements upon moving forward (and
    jumping) form it, but still allows participants to revisit previous
    pages.

    Args:
        {kwargs}

    Examples:
        Using a RevisitSection and filling it with a page in instance
        style::

            import alfred3 as al
            exp = al.Experiment()

            exp += al.RevisitSection(name="main")

            exp.main += al.Page(title="Demo", name="DemoPage")

        Using a basic section and filling it with a page in class style::

            import alfred3 as al
            exp = al.Experiment()

            @exp.member
            class Main(al.RevisitSection): pass

            @exp.member(of_section="Main")
            class DemoPage(al.Page):
                title = "Demo"
    """

    allow_forward: bool = True
    allow_backward: bool = True
    allow_jumpfrom: bool = True
    allow_jumpto: bool = True

    def _forward(self):
        super()._forward()
        self.exp.movement_manager.current_page.close()

    def _jumpfrom(self):
        super()._jumpfrom()
        self.exp.movement_manager.current_page.close()


@inherit_kwargs
class HideOnForwardSection(Section):
    """
    A section that hides pages once they have been submitted.

    Args:
        {kwargs}

    This section enables the following behaviors:

    1. Once a participant has entered their data on a page and submitted
       it by pressing "forward", the participant cannot revisit that
       page – it is hidden.
    2. The participant can, however, go back to pages in previous sections.
       For instance, they may revisit the instructions page. A press on
       "forward" will then ignore the hidden pages and take the participant
       back to the most recent page.

    .. versionadded:: 2.3.0

    Examples:

        You can test the section's behavior with this example::

            import alfred3 as al
            exp = al.Experiment()

            main = al.Section(name="main")
            main += al.Page(name="first")

            hide = al.HideOnForwardSection(name="hide")
            hide += al.Page(name="second")
            hide += al.Page(name="thirs")

            exp += main
            exp += hide

    """

    allow_forward: bool = True
    allow_backward: bool = True
    allow_jumpfrom: bool = True
    allow_jumpto: bool = True

    def _forward(self):
        super()._forward()
        self.exp.movement_manager.current_page.close()
        self.exp.movement_manager.current_page.should_be_shown = False

    def _jumpfrom(self):
        super()._jumpfrom()
        self.exp.movement_manager.current_page.close()
        self.exp.movement_manager.current_page.should_be_shown = False


@inherit_kwargs
class ForwardOnlySection(RevisitSection):
    """
    A section that allows only a single step forward; no jumping and no
    backwards steps.

    Args:
        {kwargs}

    Examples:
        Using an ForwardOnlySection and filling it with a page in instance
        style::

            import alfred3 as al
            exp = al.Experiment()

            exp += al.ForwardOnlySection(name="main")

            exp.main += al.Page(title="Demo", name="DemoPage")

        Using a basic section and filling it with a page in class style::

            import alfred3 as al
            exp = al.Experiment()

            @exp.member
            class Main(al.ForwardOnlySection): pass

            @exp.member(of_section="Main")
            class DemoPage(al.Page):
                title = "Demo"
    """

    allow_forward: bool = True
    allow_backward: bool = False
    allow_jumpfrom: bool = False
    allow_jumpto: bool = False


@inherit_kwargs
class _FinishedSection(Section):
    """
    A section that finishes the experiment on entering it.

    Args:
        {kwargs}
    """

    allow_forward: bool = False
    allow_backward: bool = False
    allow_jumpfrom: bool = False
    allow_jumpto: bool = True

    def _enter(self):
        super()._enter()
        self.experiment.finish()


class _AbortSection(Section):
    allow_forward: bool = False
    allow_backward: bool = False
    allow_jumpfrom: bool = False
    allow_jumpto: bool = True


@inherit_kwargs
class _RootSection(Section):
    """
    A section that serves as parent for all other sections in the
    experiment.

    Args:
        {kwargs}

    Defines the '_content' section and the '__finished_section' as its
    only direct children.
    """

    name = "_root"

    def __init__(self, experiment):
        super().__init__()
        self._experiment = experiment
        self.log.add_queue_logger(self, __name__)
        self.content = Section(name="_content")
        self.admin_section = None
        self.finished_section = _FinishedSection(name="__finished_section")
        self.finished_section += _DefaultFinalPage(name="_final_page")

        self._all_pages_list = None
        self._all_page_names = None

    def append_root_sections(self):
        if self.exp.admin_mode:
            from .admin import _AdminSection

            self += _AdminSection(name="_content")
            self += self.finished_section
        else:
            self += self.content
            self += self.finished_section

    @property
    def all_page_names(self):
        """
        Improvised caching mechanism for the list of all page names.

        Danger: The caching is not compatible with sections that shuffle their members.
        The shuffling does not affect this list, which is unexpected
        behavior in most cases.

        For this reason, the caching is currently deactivated.
        """

        return list(self.all_pages.keys())

        # if not self._all_page_names:
        #     self._all_page_names = list(self.all_pages.keys())

        # elif not len(self._all_page_names) == len(self.all_pages):
        #     self._all_page_names = list(self.all_pages.keys())

        # return self._all_page_names

    @property
    def all_pages_list(self):
        """
        Improvised caching mechanism for the list of all pages.

        Danger: The caching is not compatible with sections that shuffle their members.
        The shuffling does not affect this list, which is unexpected
        behavior in most cases.

        For this reason, the caching is currently deactivated.
        """

        return list(self.all_pages.values())

        # if not self._all_pages_list:
        #     self._all_pages_list = list(self.all_pages.values())

        # elif not len(self._all_pages_list) == len(self.all_pages):
        #     self._all_pages_list = list(self.all_pages.values())

        # return self._all_pages_list

    @property
    def final_page(self):
        return self.finished_section._final_page
